/* @copyright Copyright (c) 2021 Maxim Khorin <maksimovichu@gmail.com> */

'use strict';

Club.Durak = class Durak extends Club.CardPlay {

    static CARD_SELECTION_OFFSET = 16;
    static MAX_PLAYERS = 6;
    static STATUS_MOVE_DURATION = .5;

    static ACTION_ATTACK = 'attack';
    static ACTION_CONTINUE = 'continue';
    static ACTION_DEFEND = 'defend';
    static ACTION_PASS = 'pass';
    static ACTION_TRANSFER = 'transfer';

    static MESSAGE_PASS = 'Pass';
    static MESSAGE_READY = 'Ready';
    static MESSAGE_PICK_UP = 'Pick up';
    static MESSAGE_THINK = 'Thinking...';
    static MESSAGE_WAIT = '...';

    constructor () {
        super(...arguments);
        this.setGame('durak');
        this.stock = new Club.DurakStock(this);
        this.discard = new Club.DurakDiscard(this);
        this.table = new Club.DurakTable(this);
        this.events = this.createEvents();
        this.createPlayers();
        this.countdown = this.createCountdown();
        this.turnAttackStatus = this.find('.turn-status.player-attacker').get(0);
        this.turnDefenseStatus = this.find('.turn-status.player-defender').get(0);
        $(window).resize(this.onResizeWindow.bind(this));
    }

    createEvents () {
        return super.createEvents({
            BaseEvent: Club.DurakEvent,
            onHandledEvents: this.onHandledEvents.bind(this),
            onHandledHiddenEvents: this.onHandledHiddenEvents.bind(this)
        });
    }

    createCountdown () {
        return new Club.CircularCountdown({
            refreshInterval: 200,
            $element: this.master.$action
        });
    }

    start (data) {
        super.start(...arguments);
        this.master.pos = data.pos || 0;
        this.page.toggleLoading(true);
        this.events.clear();
    }

    resolveLastHiddenEventIndex () {
        const turn = this.events.getLastIndexByName(Club.DurakEvent.TURN);
        const index = turn + 1 === this.events.count()
            ? this.events.getFirstPreviousIndexByName(turn, Club.DurakEvent.DEAL)
            : turn;
        return index - 1;
    }

    startRound (data) {
        this.clear();
        this.roundData = data;
        this.round = data.round;
        this.options = data.options;
        this.winner = null;
        this.prevAttacker = null;
        this.prevDefender = null;
        this.createCards(data.deck);
        this.startPlayers(data.players);
        this.setTrump(data.trump);
        this.stock.start();
        this.toggleRoundEnd(false);
        this.arrange();
        this.startCountdown();
    }

    clear () {
        super.clear();
        this.clearCards();
        this.stock.clear();
        this.table.clear();
        this.discard.clear();
        this.players.forEach(player => player.clear());
        this.countdown.clear();
    }

    startCountdown () {
        const timeout = this.options?.actionTimeout * 1000;
        this.countdown.start(timeout || 0, this.messageTimestamp);
    }

    startPlayers (items) {
        this.playerMap = {};
        for (let i = 0; i < items.length; ++i) {
            const pos = (i + this.master.pos) % items.length;
            this.playerMap[pos] = this.players[i];
            this.players[i].start(pos, items[pos]);
        }
        this.setDataAttr('opponents', items.length - 1);
    }

    getPlayer (pos) {
        return this.playerMap[pos];
    }

    getActiveAttacker () {
        if (!this.options.siege) {
            return this.isReadyToAttack(this.attacker) ? this.attacker : null;
        }
        const counter = this.countRoundPlayers();
        for (let i = 0; i < counter; ++i) {
            const player = this.getPlayer((this.attacker.pos + i) % counter);
            if (this.isReadyToAttack(player)) {
                return player;
            }
        }
    }

    createPlayers () {
        this.master = new Club.DurakMaster(0, this);
        this.players = [this.master];
        this.playerMap = {};
        const $opponents = this.find('.opponents');
        for (let num = 1; num < Club.Durak.MAX_PLAYERS; ++num) {
            $opponents.append(this.resolveTemplate('opponent', {num}));
            this.players.push(new Club.DurakOpponent(num, this));
        }
        Jam.t($opponents);
    }

    countRoundPlayers () {
        return this.roundData.players.length;
    }

    canBeat (attacking, defending) {
        if (attacking.getSuit() !== defending.getSuit()) {
            return this.isTrump(defending);
        }
        if (attacking.getRank() < defending.getRank()) {
            return true;
        }
        return this.options.fallenAce
            && attacking.getRank() === this.options.maxRank
            && defending.getRank() === this.options.minRank
            && this.isTrump(attacking);
    }

    canCardAttack (card) {
        if (this.table.isEmpty()) {
            return true;
        }
        if (this.isAttackLimit()) {
            return false;
        }
        return this.options.siege
            ? this.table.isSameRank(card)
            : this.table.isSameAttackingRank(card);
    }

    canTransfer (card) {
        return this.options.transferable
            && this.table.countAttacks() < this.getMaxAttacks()
            && this.table.canTransfer(card)
            && this.getTransferTarget();
    }

    getTransferTarget () {
        const counter = this.countRoundPlayers();
        for (let i = 1; i < counter; ++i) {
            const player = this.getPlayer((this.defender.pos + i) % counter);
            if (player.cards.count() > 0) {
                return player.cards.count() > this.table.countAttacks() ? player : null;
            }
        }
    }

    getMaxAttacks () {
        return this.discard.count()
            ? this.options.maxAttacks
            : this.options.maxAttacksBeforeDiscard;
    }

    isAttackLimit () {
        return this.table.countAttacks() === this.getMaxAttacks()
            || this.table.getOpenAttackingCards().length === this.defender.cards.count()
            || this.table.hasOnlyFullRanks();
    }

    isFinished () {
        return this.events.getLastIndexByName(Club.DurakEvent.END) !== undefined;
    }

    isReadyToAttack (player) {
        return !player.turned && player !== this.defender && player.cards.count();
    }

    isTrump (card) {
        return card.getSuit() === this.trump?.suit;
    }

    setTrump (data) {
        this.trump = data;
        this.setDataAttr('trump', data?.suit);
    }

    setAttacker (player) {
        this.prevAttacker = this.attacker;
        this.setDataAttr('attacker', player ? player.num : null);
        this.attacker = player;
    }

    setDefender (player) {
        this.prevDefender = this.defender;
        this.setDataAttr('defender', player ? player.num : null);
        this.defender = player;
    }

    resolveWinner (player) {
        if (!player || !player.isEmpty() || !this.stock.isEmpty()) {
            return false;
        }
        if (!this.winner) {
            player.setWinner();
            this.winner = player;
            return true;
        }
        if (this.players.filter(player => !player.isEmpty()).length > 1) {
            player.setFinisher();
        }
    }

    onMessage (data) {
        super.onMessage(data);
        this.events.add(data.events);
        this.events.process();
        this.startCountdown();
    }

    onHandledHiddenEvents () {
        this.page.toggleLoading(false);
        this.toggleClass('hidden', false);
        this.arrange();
    }

    onHandledEvents () {
        if (!this.events.prediction) {
            this.master.update();
        }
    }

    arrange () {
        this.setCardSize();
        this.stock.arrange();
        this.table.arrange();
        this.discard.arrange();
        this.players.forEach(player => player.arrange());
    }

    updateTurnedPlayers () {
        this.players.forEach(player => player.setTurned(player.isEmpty()));
    }

    updatePlayerMessages () {
        this.players.forEach(player => player.updateMessage());
    }

    toggleRoundEnd (state) {
        this.finished = state;
        this.toggleClass('end', state);
        if (state) {
            this.setAttacker(null);
            this.setDefender(null);
        }
    }

    onResizeWindow () {
        if (!this.motion.isActive()) {
            this.arrange();
        }
    }

    showTurn () {
        if (!this.prevAttacker) {
            return null;
        }
        this.toggleClass('show-statuses', true); 
        if (this.attacker !== this.prevAttacker) {
            this.moveStatus(this.turnAttackStatus, this.attacker, this.prevAttacker);
        }
        if (this.defender !== this.prevDefender) {
            this.moveStatus(this.turnDefenseStatus, this.defender, this.prevDefender);
        }
        this.toggleClass('show-statuses', false);
        this.toggleClass('turning', true);
        this.motion.done(() => this.toggleClass('turning', false));
    }

    moveCard (card) {
        card.toggleClass('selected', false);
        return super.moveCard(...arguments);
    }

    moveStatus (status, player, prevPlayer) {
        this.motion.move({
            element: status,
            source: prevPlayer.getStatusOffset(),
            target: player.getStatusOffset(),
            duration: Club.Durak.STATUS_MOVE_DURATION
        });
    }

    hasExportData () {
        return !this.isFinished();
    }

    exportData () {
        return Object.assign(super.exportData(), {
            events: this.events.items,
            pos: this.master.pos
        });
    }
};
Club.DurakEvent = class DurakEvent {

    static ATTACK = 'attack';
    static DEAL = 'deal';
    static DEFEND = 'defend';
    static DISCARD = 'discard';
    static END = 'end';
    static PASS = 'pass';
    static READY = 'ready';
    static ROUND = 'round';
    static PICK_UP = 'pickUp';
    static TRANSFER = 'transfer';
    static TURN = 'turn';

    static getClass (name) {
        switch (name) {
            case this.ATTACK: return Club.DurakAttackEvent;
            case this.DEAL: return Club.DurakDealEvent;
            case this.DEFEND: return Club.DurakDefendEvent;
            case this.DISCARD: return Club.DurakDiscardEvent;
            case this.END: return Club.DurakEndEvent;
            case this.PASS: return Club.DurakPassEvent;
            case this.READY: return Club.DurakReadyEvent;
            case this.ROUND: return Club.DurakRoundEvent;
            case this.PICK_UP: return Club.DurakPickUpEvent;
            case this.TRANSFER: return Club.DurakTransferEvent;
            case this.TURN: return Club.DurakTurnEvent;
        }
    }

    constructor (params) {
        this.play = params.play;
        this.table = this.play.table;
        this.hidden = params.hidden;
        this.data = params.data;
        this.player = this.getPlayer(this.data?.[0]);
        this.index = params.index;
        this.onHandled = params.onHandled;
    }

    isMaster () {
        return this.player === this.play.master;
    }

    isMasterPassInPreviousEvents () {
        const items = this.play.events.slice(0, this.index).reverse();
        for (const [name, data] of items) {
            if (name !== Club.DurakEvent.PASS) {
                return false;
            }
            if (this.play.master.pos === data[0]) {
                return true;
            }
        }
    }

    getPlayer (pos) {
        return this.play.getPlayer(pos);
    }

    process () {
        this.hidden ? this.processHidden() : this.processNormal();
    }

    processHidden () {
        this.finish();
    }

    processNormal () {
        this.finish();
    }

    finish () {
        this.onHandled(this);
    }

    finishAfterMotion (delay = 0) {
        this.play.motion.done(() => setTimeout(this.finish.bind(this), delay));
    }

    arrangeAfterMotion (player) {
        this.play.motion.done(player.arrange.bind(player));
    }

    openCard (card, index) {
        card.open(this.items[index].rank, this.items[index].suit);
        card.toggleClass('selected', false);
    }

    closeCard (card) {
        if (!this.play.keepFacedCards) {
            card.close()
        }
    }
};
Club.DurakPlayer = class DurakPlayer {

    constructor (num, play) {
        this.play = play;
        this.num = num;
        this.$container = this.getContainer();
        this.cards = this.createCardList();
        this.$won = this.$container.find('.player-stat-won');
        this.$lost = this.$container.find('.player-stat-lost');
        this.$drawn = this.$container.find('.player-stat-drawn');
        this.clear();
    }

    isEmpty () {
        return this.cards.isEmpty();
    }

    isAttacker () {
        return this.play.attacker === this;
    }

    isDefender () {
        return this.play.defender === this;
    }

    isRoundEnd () {
        return this.play.finished;
    }

    activate () {
        this.active = true;
    }

    deactivate () {
        this.active = false;
    }

    setTurned (state) {
        this.turned = state;
    }

    toggleClass () {
        this.$container.toggleClass(...arguments);
    }

    updateMessage () {
        this.setMessage(this.getMessage());
    }

    getMessage () {
        if (this.isDefender()) {
            if (!this.play.table.hasOpenAttack()) {
                return Club.Durak.MESSAGE_WAIT;
            }
            if (this.turned) {
                return Club.Durak.MESSAGE_PICK_UP;
            }
            return this.play.getActiveAttacker()
                ? Club.Durak.MESSAGE_WAIT
                : Club.Durak.MESSAGE_THINK;
        }
        if (this.play.isAttackLimit()) {
            return Club.Durak.MESSAGE_WAIT;
        }
        if (this.play.getActiveAttacker() === this) {
            return Club.Durak.MESSAGE_THINK;
        }
        return !this.turned || this.isEmpty()
            ? Club.Durak.MESSAGE_WAIT
            : Club.Durak.MESSAGE_PASS;
    }

    setMessage () {}

    getContainer () {}

    clear () {
        this.cards.clear();
        this.active = false;
        this.setEndStatus('');
        this.setMessage('');
    }

    createCardList () {
        return this.play.createCardList(...arguments);
    }

    start (pos, data) {
        this.pos = pos;
        this.data = data;
        this.$won.html(data.won);
        this.$lost.html(data.lost);
        this.$drawn.html(data.drawn);
        this.setMessage(Club.Durak.MESSAGE_WAIT);
        this.setName();
        this.arrange();
    }

    setName () {
        const name = this.play.resolvePlayerName(this.data);
        this.$container.find('.player-name').html(name).attr('title', name);
    }

    onMoveCard (card, data) {
        this.cards.add(card);
        if (data) {
            card.open(data.rank, data.suit);
        }
    }

    removeCard () {
        this.cards.remove(...arguments);
    }

    getRightSidePlayerLastCard () {
        return null;
    }

    openCards () {}

    setWinner () {
        this.setEndStatus('winner');
        this.incrementStatCounter(this.$won);
    }

    setFinisher () {
        this.setEndStatus('out');
    }

    setLoser (cards) {
        this.openCards(cards);
        this.setEndStatus('loser');
        this.incrementStatCounter(this.$lost);
    }

    setDraw () {
        this.setEndStatus('draw');
        this.incrementStatCounter(this.$drawn);
    }

    setEndStatus (status) {
        this.endStatus = status;
        this.$container.attr('data-end', status);
    }

    incrementStatCounter ($stat) {
        $stat.html(parseInt($stat.html()) + 1);
    }

    getStatusOffset () {
        return this.play.getOffset(this.$container.find('.player-status').first());
    }
};
Club.DurakDiscard = class DurakDiscard {

    constructor (play) {
        this.play = play;
        this.cards = play.createCardList();
    }

    getCardOffset () {
        return [this.rect.x, this.rect.y];
    }

    clear () {
        this.cards.clear();
    }

    count () {
        return this.cards.count();
    }

    arrange () {
        this.rect = this.play.getElementRect('.discard');
        this.cards.setOffset(this.rect.x, this.rect.x, this.rect.w);
        this.cards.arrange();
    }
};
Club.DurakAttackEvent = class DurakAttackEvent extends Club.DurakEvent {

    static PROCESS_DELAY = 100;

    constructor () {
        super(...arguments);
        this.items = this.data[1];
    }

    processNormal () {
        this.table.arrange(this.items.length);
        const cards = this.execute();
        cards.forEach(this.arrangeCard, this);
        setTimeout(this.processCards.bind(this, cards), this.constructor.PROCESS_DELAY);
    }

    arrangeCard (card) {
        if (!this.isMaster()) {
            this.closeCard(card);
        }
        const index = this.table.getAttackingIndex(card);
        this.table.getDefendingCard(index - 1)?.after(card);
    }

    processCards (cards) {
        cards.forEach(this.processCard, this);
        this.play.motion.done(this.onMotionDone.bind(this));
    }

    processCard (card, index) {
        const offset = this.table.getAttackingOffset(card);
        this.play.moveCard(card, offset).done(() => this.openCard(card, index));
    }

    onMotionDone () {
        this.player.arrange();
        this.play.table.arrange();
        this.finishAfterMotion();
    }

    processHidden () {
        this.execute();
        this.finish();
    }

    execute () {
        const attacker = this.player;
        const cards = [];
        for (const {rank, suit} of this.items) {
            let card = attacker.cards.find(rank, suit);
            if (!card && !this.isMaster()) {
                card = attacker.cards.last();
                card.open(rank, suit);
            }
            if (card) {
                attacker.removeCard(card);
                this.table.addAttack(card);
                cards.push(card);
            }
        }
        this.play.isAttackLimit()
            ? this.play.players.forEach(player => player.setTurned(true))
            : attacker.setTurned(attacker.isEmpty());
        this.play.defender.setTurned(false);
        this.play.updatePlayerMessages();
        this.play.resolveWinner(attacker);
        return cards;
    }
};
Club.DurakDealEvent = class DurakDealEvent extends Club.DurakEvent {

    static MOVE_DELAY = .1;
    static MOVE_DURATION = .25;

    constructor () {
        super(...arguments);
        this.deals = this.createDeals();
    }

    createDeals () {
        return this.data.map(this.createDeal, this);
    }

    createDeal ([pos, items]) {
        const total = Array.isArray(items) ? items.length : items;
        const player = this.getPlayer(pos);
        const opened = Array.isArray(items);
        return {total, player, opened, items};
    }

    processNormal () {
        for (const {player, total} of this.deals) {
            player.cards.arrange(total);
        }
        const items = this.getMoveItems();
        this.execute();
        items.forEach(this.processMoveItem, this);
        this.finishAfterMotion();
    }

    getMoveItems () {
        const result = [];
        const {cardsDealtAtOnce, defaultCardsInHand} = this.play.options;
        const turns = Math.ceil(defaultCardsInHand / cardsDealtAtOnce);
        for (let turn = 0; turn < turns; ++turn) {
            for (const {player, total, items} of this.deals) {
                const start = turn * cardsDealtAtOnce;
                for (let i = start; i < total && i < start + cardsDealtAtOnce; ++i) {
                    const card = this.play.stock.splice(1)[0];
                    result.push([card, items[i], player]);
                    player.cards.add(card);
                }
            }
        }
        return result;
    }

    processMoveItem ([card, data, player]) {
        const offset = player.cards.getCardOffset(card);
        this.play.moveCard(card, offset, this.constructor.MOVE_DURATION, this.constructor.MOVE_DELAY)
            .done(this.afterCardMove.bind(this, ...arguments));
    }

    afterCardMove ([card, data, player]) {
        data ? this.openCard(card, data)
             : this.closeCard(card);
        player.cards.setCardOrder(card, player.getRightSidePlayerLastCard());
    }

    processHidden () {
        this.execute();
        for (const {player, total, items} of this.deals) {
            const cards = this.play.stock.splice(total);
            player.cards.add(cards);
            Array.isArray(items)
                ? cards.forEach((card, index) => this.openCard(card, items[index]))
                : cards.forEach(this.closeCard, this);
        }
        this.finish();
    }

    execute () {
        for (const {player} of this.deals) {
            player.setTurned(false);
            player.setMessage(Club.Durak.MESSAGE_WAIT);
        }
    }

    openCard (card, data) {
        card.open(data.rank, data.suit);
    }
};
Club.DurakDefendEvent = class DurakDefendEvent extends Club.DurakEvent {

    static PROCESS_DELAY = 100;

    constructor () {
        super(...arguments);
        this.items = this.data[1];
    }

    processNormal () {
        const pairs = this.execute();
        pairs.forEach(({attacking, defending}) => attacking.after(defending));
        setTimeout(this.processPairs.bind(this, pairs), this.constructor.PROCESS_DELAY);
    }

    processPairs (pairs) {
        pairs.forEach(this.processPair, this);
        this.play.motion.done(this.onMotionDone.bind(this));
    }

    processPair ({attacking, defending}, index) {
        const offset = this.table.getDefendingOffset(defending);
        this.play.moveCard(defending, offset)
            .done(() => this.openCard(defending, index));
    }

    onMotionDone () {
        this.player.arrange();
        this.play.table.arrange();
        this.finishAfterMotion();
    }

    processHidden () {
        this.execute().forEach(({defending}, index) => this.openCard(defending, index));
        this.finish();
    }

    execute () {
        const defender = this.player;
        const cards = [];
        for (let [attacking, defending] of this.items) {
            attacking = this.play.cards.find(attacking.rank, attacking.suit);
            defending = defender.cards.find(defending.rank, defending.suit);
            if (!defending && !this.isMaster()) {
                defending = defender.cards.last();
            }
            if (defending) {
                defender.removeCard(defending);
                this.table.addDefense(attacking, defending);
                cards.push({attacking, defending});
            }
        }
        if (!this.table.hasOpenAttack()) {
            this.play.updateTurnedPlayers();
        }
        this.play.updatePlayerMessages();
        this.play.resolveWinner(defender);
        return cards;
    }

    openCard (card, index) {
        card.open(this.items[index][1].rank, this.items[index][1].suit);
    }
};
Club.DurakDiscardEvent = class DurakDiscardEvent extends Club.DurakEvent {

    static AFTER_MASTER_DELAY = 10;
    static AFTER_OPPONENT_DELAY = 1000;
    static MOVE_DELAY = .1;
    static MOVE_DURATION = .2;

    constructor () {
        super(...arguments);
        this.cards = this.table.getCards();
    }

    processNormal () {
        setTimeout(() => {
            this.execute().reverse().forEach(this.processCard, this);
            this.finishAfterMotion();
        }, this.getProcessDelay());
    }

    processCard (card) {
        const offset = this.play.discard.getCardOffset(card);
        this.play.moveCard(card, offset, this.constructor.MOVE_DURATION, this.constructor.MOVE_DELAY)
            .done(this.closeCard.bind(this, card));
    }

    getProcessDelay () {
        return this.play.master.isDefender() || this.isMasterPassInPreviousEvents()
            ? this.constructor.AFTER_MASTER_DELAY
            : this.constructor.AFTER_OPPONENT_DELAY;
    }

    processHidden () {
        this.execute();
        this.finish();
    }

    execute () {
        this.play.discard.cards.add(this.cards);
        this.table.clear();
        return this.cards;
    }
};
Club.DurakEndEvent = class DurakEndEvent extends Club.DurakEvent {

    process () {
        const play = this.play;
        if (play.master !== play.winner) {
            play.master.setFinisher();
        }
        const data = this.data;
        const winner = this.getPlayer(data.winner);
        const loser = this.getPlayer(data.loser);
        const draw1 = this.getPlayer(data.draw1);
        const draw2 = this.getPlayer(data.draw2);
        loser?.setLoser(data.loserCards);
        draw1?.setDraw();
        draw2?.setDraw();
        play.resolveWinner(winner);
        play.players.forEach(this.togglePlayerEnd, this);
        play.master.activate();
        play.master.toggleAction(true);
        play.toggleRoundEnd(true);
        this.finish();
    }

    togglePlayerEnd (player) {
        player.setTurned(false);
        player.setMessage(Club.Durak.MESSAGE_THINK);
    }
};
Club.DurakPassEvent = class DurakPassEvent extends Club.DurakEvent {

    process () {
        this.player.setTurned(true);
        this.play.updatePlayerMessages();
        this.finish();
    }
};
Club.DurakPickUpEvent = class DurakPickUpEvent extends Club.DurakEvent {

    static AFTER_MASTER_DELAY = 10;
    static AFTER_OPPONENT_DELAY = 1000;
    static MOVE_DELAY = .1;
    static MOVE_DURATION = .2;

    constructor () {
        super(...arguments);
        this.cards = this.table.getCards().reverse();
    }

    processNormal () {
        setTimeout(() => {
            this.playerCards = this.player.cards;
            this.playerCards.arrange(this.cards.length);
            this.rightPlayerCard = this.player.getRightSidePlayerLastCard();
            this.execute().forEach(this.processCard, this);
            this.finishAfterMotion();
        }, this.getProcessDelay());
    }

    processCard (card) {
        const offset = this.playerCards.getCardOffset(card);
        this.play.moveCard(card, offset, this.constructor.MOVE_DURATION, this.constructor.MOVE_DELAY)
            .done(this.afterCardMove.bind(this, card));
    }

    afterCardMove (card) {
        if (!this.isMaster()) {
            this.closeCard(card);
        }
        this.playerCards.setCardOrder(card, this.rightPlayerCard);
    }

    getProcessDelay () {
        return this.isMaster() || this.isMasterPassInPreviousEvents()
            ? this.constructor.AFTER_MASTER_DELAY
            : this.constructor.AFTER_OPPONENT_DELAY;
    }

    processHidden () {
        const cards = this.execute();
        if (!this.isMaster()) {
            cards.forEach(this.closeCard, this);
        }
        this.finish();
    }

    execute () {
        this.player.cards.add(this.cards);
        this.table.clear();
        return this.cards;
    }
};
Club.DurakReadyEvent = class DurakReadyEvent extends Club.DurakEvent {

    process () {
        this.player.deactivate();
        this.player.setTurned(true);
        this.player.setMessage(Club.Durak.MESSAGE_READY);
        this.finish();
    }
};
Club.DurakRoundEvent = class DurakRoundEvent extends Club.DurakEvent {

    process () {
        this.play.events.clearBeforeCursor();
        this.play.events.hiddenIndex = this.play.resolveLastHiddenEventIndex();
        this.play.startRound(this.data);
        this.finish();
    }
};
Club.DurakTransferEvent = class DurakTransferEvent extends Club.DurakEvent {

    static PROCESS_DELAY = 250;

    constructor () {
        super(...arguments);
        this.from = this.play.defender;
        this.target = this.player;
        this.items = this.data[1];
    }

    processNormal () {
        this.table.arrange(this.items.length);
        this.executeCards().forEach(this.processCard, this);
        this.play.motion.done(this.processTurn.bind(this));
    }

    processCard (card, index) {
        if (this.from !== this.play.master) {
            this.closeCard(card);
        }
        const offset = this.table.getAttackingOffset(card);
        this.play.moveCard(card, offset).done(() => this.openCard(card, index));
    }

    processTurn () {
        this.from.arrange();
        this.executeTurn();
        this.play.showTurn();
        this.finishAfterMotion(this.constructor.PROCESS_DELAY);
    }

    processHidden () {
        this.executeCards();
        this.executeTurn();
        this.finish();
    }

    executeCards () {
        const cards = [];
        for (const {rank, suit} of this.items) {
            let card = this.from.cards.find(rank, suit);
            if (!card && this.from !== this.play.master) {
                card = this.from.cards.last();
                card.open(rank, suit);
            }
            if (card) {
                this.from.removeCard(card);
                this.table.addAttack(card);
                cards.push(card);
            }
        }
        return cards;
    }

    executeTurn () {
        const play = this.play;
        play.setAttacker(this.from);
        play.setDefender(this.target);
        play.isAttackLimit()
            ? play.players.forEach(player => player.setTurned(true))
            : play.updateTurnedPlayers();
        play.defender.setTurned(false);
        play.updatePlayerMessages();
        play.resolveWinner(this.from);
    }
};
Club.DurakTurnEvent = class DurakTurnEvent extends Club.DurakEvent {

    static PROCESS_DELAY = 250;

    processNormal () {
        this.execute();
        this.play.showTurn();
        this.finishAfterMotion(this.constructor.PROCESS_DELAY);
    }

    processHidden () {
        this.execute();
        this.finish();
    }

    execute () {
        this.play.setAttacker(this.getPlayer(this.data.attacker));
        this.play.setDefender(this.getPlayer(this.data.defender));
        this.play.updateTurnedPlayers();
        this.play.updatePlayerMessages();
    }
};
Club.DurakPlayback = class DurakPlayback extends Club.Durak {

    constructor () {
        super(...arguments);
        this.on('click', '.playback-action-pause', this.onPause.bind(this));
        this.on('click', '.playback-action-play', this.onPlay.bind(this));
        this.on('click', '.playback-action-start', this.onStart.bind(this));
        this.on('click', '.playback-action-end', this.onEnd.bind(this));
        this.keepFacedCards = true;
        this.toggleClass('playback', true);
    }

    onHandledHiddenEvents () {
        this.arrange();
    }

    onHandledEvents () {
        this.togglePause(true);
    }

    start (data) {
        super.start(...arguments);
        this.page.toggleLoading(false);
        this.toggleClass('hidden', false);
        this.events.add(data.events);
        this.processEvents();
    }

    processEvents () {
        this.events.stopIndex = 2;
        this.events.process();
    }

    startPlayback () {
        this.events.process();
    }

    onPlay () {
        this.togglePause(false);
        this.events.isEnded()
            ? this.onStart()
            : this.events.continue();
    }

    onPause () {
        this.togglePause(true);
        this.events.stop();
    }

    onStart () {
        this.togglePause(false);
        this.events.reset();
        this.processEvents();
    }

    onEnd () {
        this.events.hiddenIndex = this.events.count() - 1;
        this.onPlay();
    }

    togglePause () {
        this.toggleClass('paused', ...arguments);
    }

    attachSocket () {}

    startCountdown () {}

    resolveLastHiddenEventIndex () {
        return -1;
    }
};
Club.DurakMaster = class DurakMaster extends Club.DurakPlayer {

    static AMBIGUOUS_MESSAGE = 'This is an ambiguous defense. Choose one attacking card';
    static CARD_BOTTOM_FACTOR = 1.8;
    static CARD_HELP = 'Press on the top of a card to action. Press the bottom to toggle selection';
    static CARD_HELP_KEY = 'club.durak.cardHelp';

    constructor () {
        super(...arguments);
        this.play.on('click', '.card[data-rank][data-suit]', this.onCard.bind(this));
        this.play.on('click', '.master-cards .small-card', this.onSmallCard.bind(this));
        this.play.on('click', '.master-scale', this.onScale.bind(this));
        this.$action = this.$container.find('.master-action');
        this.$action.click(this.onAction.bind(this));
    }

    getContainer () {
        return this.play.find(`.master`);
    }

    update () {
        this.canAct() ? this.activate() : this.deactivate();
    }

    canAct () {
        if (this.isRoundEnd()) {
            return !this.turned;
        }
        if (this.turned || this.isEmpty()) {
            return false;
        }
        const table = this.play.table;
        if (this.isDefender()) {
            return table.hasOpenAttack() && !this.play.getActiveAttacker();
        }
        if (table.isEmpty()) {
            return this.isAttacker();
        }
        if (this.play.isAttackLimit()) {
            return false;
        }
        return this.play.getActiveAttacker() === this;
    }

    activate () {
        super.activate();
        this.play.toggleClass('master', true);
        this.play.toggleClass('defender', this.isDefender());
        this.toggleAction(this.isRoundEnd() || !this.play.table.isEmpty());
    }

    deactivate () {
        super.deactivate();
        this.play.toggleClass('master', false);
        this.toggleAction(false);
        this.toggleScale(false);
        this.resetSelection();
    }

    onAction () {
        this.resetSelection();
        this.isRoundEnd() ? this.continue() : this.pass();
    }

    continue () {
        this.send(Club.Durak.ACTION_CONTINUE);
    }

    pass () {
        this.send(Club.Durak.ACTION_PASS);
    }

    onCard (event) {
        const target = $(event.target).data('card');
        if (!this.active || !target || this.isRoundEnd()) {
            return false;
        }
        if (this.selection && this.play.table.isOpenAttackingCard(target)) {
            return this.move(target);
        }
        if (!this.cards.has(target)) {
            return false;
        }
        const bottom = event.offsetY * this.constructor.CARD_BOTTOM_FACTOR > this.play.getCardHeight();
        if (bottom) {
            this.showCardHelp();
        }
        if (this.selection === target) {
            return bottom
                ? this.resetSelection()
                : this.move(null, true);
        }
        if (bottom) {
            return this.setSelection(target);
        }
        this.setSelection(target);
        this.move();
    }

    move (target, reselection) {
        const card = this.selection;
        if (!card) {
            return false;
        }
        if (!this.isDefender()) {
            return this.attack();
        }
        if (target) {
            return this.defend(target);
        }
        const cards = this.play.table.getWeakerAttackingCards(card);
        if (this.play.canTransfer(card)) {
            return !cards.length || reselection ? this.transfer() : null;
        }
        if (cards.length === 1) {
            return this.defend(cards[0]);
        }
        if (cards.length && reselection) {
            this.showMessage(this.constructor.AMBIGUOUS_MESSAGE);
        }
    }

    attack () {
        if (this.play.getActiveAttacker() === this && this.play.canCardAttack(this.selection)) {
            const cards = [this.selection.getData()];
            this.addPredictedEvent(Club.DurakEvent.ATTACK, [this.pos, cards]);
            this.send(Club.Durak.ACTION_ATTACK, {cards});
            return true;
        }
    }

    defend (target) {
        if (this.play.canBeat(target, this.selection)) {
            const pairs = [[target.getData(), this.selection.getData()]];
            this.addPredictedEvent(Club.DurakEvent.DEFEND, [this.pos, pairs]);
            this.send(Club.Durak.ACTION_DEFEND, {pairs});
            return true;
        }
    }

    transfer () {
        const cards = [this.selection.getData()];
        const target = this.play.getTransferTarget();
        this.addPredictedEvent(Club.DurakEvent.TRANSFER, [target.pos, cards]);
        this.send(Club.Durak.ACTION_TRANSFER, {cards});
        return true;
    }

    toggleAction (state) {
        this.$action.toggleClass('active', state);
    }

    setSelection (card) {
        this.resetSelection();
        this.selection = card;
        this.arrangeSelection();
    }

    resetSelection () {
        this.resetSelectionOffset();
        this.toggleSelectedClass(false);
        this.selection = null;
    }

    resetSelectionOffset () {
        this.selection?.changeOffset(0, Club.Durak.CARD_SELECTION_OFFSET);
    }

    toggleSelectedClass (state) {
        this.selection?.toggleClass('selected', state);
    }

    removeCard (card) {
        this.selection = null;
        return super.removeCard(card);
    }

    setEndStatus (status) {
        super.setEndStatus(status);
        this.play.find('.round-message').attr('data-status', status);
    }

    arrange () {
        const rect = this.play.getElementRect('.master-cards');
        this.cards.setOffset(rect.x, rect.y, rect.w);
        this.cards.arrange();
        this.arrangeSelection();
    }

    arrangeSelection () {
        this.selection?.changeOffset(0, -Club.Durak.CARD_SELECTION_OFFSET);
        this.toggleSelectedClass(true);
    }

    send () {
        this.selection = null;
        this.deactivate();
        this.play.send(...arguments);
    }

    showCardHelp () {
        if (!Jam.localStorage.get(this.constructor.CARD_HELP_KEY)) {
            this.showMessage(this.constructor.CARD_HELP);
            Jam.localStorage.set(this.constructor.CARD_HELP_KEY, true);
        }
    }

    showMessage (message) {
        return Jam.dialog.info(this.play.translate(message), {strictCancel: true});
    }

    onSmallCard (event) {
        const {rank, suit} = $(event.target).data();
        const card = this.cards.find(rank, suit);
        this.toggleScale(false);
        this.setSelection(card);
    }

    onScale () {
        this.resetSelection();
        this.toggleScale();
        this.createSmallCards();
    }

    toggleScale () {
        this.play.toggleClass('master-scaled', ...arguments);
    }

    createSmallCards () {
        const $container = this.play.find('.master-cards').empty();
        for (const card of this.cards) {
            $container.append(this.createSmallCard(card));
        }
    }

    createSmallCard (card) {
        const element = document.createElement('div');
        element.setAttribute('data-rank', card.getRank());
        element.setAttribute('data-suit', card.getSuit());
        element.classList.add('small-card');
        return element;
    }

    addPredictedEvent () {
        this.play.events.addPrediction(...arguments);
        this.play.events.process();
    }
};
Club.DurakOpponent = class DurakOpponent extends Club.DurakPlayer {

    getContainer () {
        return this.play.find(`.opponent-${this.num}`);
    }

    createCardList () {
        this.$cards = this.$container.find('.player-cards');
        this.$message = this.$container.find('.player-message');
        this.$counter = this.$container.find('.player-card-counter');
        this.$counter.on('cards:change', this.onChangeCards.bind(this));
        return super.createCardList({
            $listener: this.$counter
        });
    }

    start () {
        super.start(...arguments);
        this.$container.attr('data-type', this.data.type);
    }

    arrange () {
        const rect = Club.getElementRect(this.$container);
        const list = Club.getElementRect(this.$cards);
        this.cards.setCardOrder(this.cards.get(0), this.getRightSidePlayerLastCard());
        this.cards.setOffset(rect.x + list.x, rect.y + list.y, list.w);
        this.cards.arrange();
    }

    setMessage (message) {
        this.$message.html(this.play.translate(message));
    }

    onChangeCards () {
        const total = this.cards.count();
        this.$counter.html(total).toggleClass('empty', total === 0);
    }

    getRightSidePlayerLastCard () {
        return this.num > 1
            ? this.play.players[this.num - 1].cards.last()
            : null;
    }

    openCards (data) {
        data.forEach(({rank, suit}, index) => this.cards.get(index).open(rank, suit));
    }
};
Club.DurakStock = class DurakStock {

    constructor (play) {
        this.play = play;
        this.cards = play.createCardList();
    }

    isEmpty () {
        return !this.trumpCard && this.cards.isEmpty();
    }

    start () {
        this.clear();
        this.cards.add(this.play.cards);
        this.cards.close();
        this.setTrumpCard(this.play.trump);
    }

    setTrumpCard (trump) {
        if (trump) {
            this.trumpCard = this.cards.get(0);
            this.trumpCard.open(trump.rank, trump.suit);
            this.cards.remove(this.trumpCard);
        }
    }

    clear () {
        this.cards.clear();
        this.trumpCard = null;
    }

    arrange () {
        const rect = this.play.getElementRect('.stock');
        const pile = this.play.getElementRect('.stock-pile');
        this.cards.setOffset(rect.x + pile.x, rect.y + pile.y, pile.w);
        this.cards.arrange();
        if (this.trumpCard) {
            this.trumpCard.setOffset(rect.x, rect.y);
        }
    }

    splice (amount) {
        const index = this.cards.count() < amount ? amount - 1 : amount;
        const cards = this.cards.splice(index).reverse();
        if (cards.length < amount) {
            cards.push(this.trumpCard);
            this.trumpCard = null;
        }
        return cards;
    }
};
Club.DurakTable = class DurakTable {

    static MAX_CARD_SPACE = 10;

    constructor (play) {
        this.play = play;
        this.clear();
    }

    isEmpty () {
        return !this.countAttacks();
    }

    isSameRank (card) {
        const rank = card.getRank();
        for (const [attacking, defending] of this.pairs) {
            if (attacking.getRank() === rank || (defending?.getRank() === rank)) {
                return true;
            }
        }
    }

    isSameAttackingRank (card) {
        const rank = card.getRank();
        for (const [attacking] of this.pairs) {
            if (attacking.getRank() === rank) {
                return true;
            }
        }
    }

    isOpenAttackingCard (card) {
        for (const [attacking, defending] of this.pairs) {
            if (card === attacking && !defending) {
                return true;
            }
        }
    }

    hasOnlyFullRanks () {
        if (this.pairs.length !== 4) {
            return false;
        }
        const attackingRank = this.pairs[0][0]?.getRank();
        for (const [attacking] of this.pairs) {
            if (attacking.getRank() !== attackingRank) {
                return false;
            }
        }
        const defendingRank = this.pairs[0][1]?.getRank();
        for (const [, defending] of this.pairs) {
            if (defending?.getRank() !== defendingRank) {
                return false;
            }
        }
        return true;
    }

    hasOpenAttack () {
        for (const pair of this.pairs) {
            if (!pair[1]) {
                return true;
            }
        }
    }

    addAttack (attacking) {
        this.pairs.push([attacking]);
    }

    addDefense (attacking, defending) {
        const index = this.getAttackingIndex(attacking);
        this.pairs[index][1] = defending;
    }

    canTransfer (card) {
        const rank = card.getRank();
        for (const [attacking, defending] of this.pairs) {
            if (defending || attacking.getRank() !== rank) {
                return false;
            }
        }
        return this.pairs.length > 0;
    }

    clear () {
        this.pairs = [];
    }

    countAttacks () {
        return this.pairs.length;
    }

    getCards () {
        const cards = [];
        for (const [attacking, defending] of this.pairs) {
            cards.push(attacking);
            if (defending) {
                cards.push(defending);
            }
        }
        return cards;
    }

    getLastCard () {
        const pair = this.pairs[this.pairs.length - 1];
        return pair ? (pair[1] || pair[0]) : null;
    }

    getAttackingCard (index) {
        return this.pairs[index]?.[0];
    }

    getDefendingCard (index) {
        return this.pairs[index]?.[1];
    }

    getWeakerAttackingCards (defending) {
        const cards = [];
        for (const attacking of this.getOpenAttackingCards()) {
            if (this.play.canBeat(attacking, defending)) {
                cards.push(attacking);
            }
        }
        return cards;
    }

    getOpenAttackingCards () {
        const cards = [];
        for (const [attacking, defending] of this.pairs) {
            if (!defending) {
                cards.push(attacking);
            }
        }
        return cards;
    }

    getAttackingOffset (card) {
        const [dx, dy] = this.getOffsetStep(this.countAttacks());
        const index = this.getAttackingIndex(card);
        return [
            this.rect.x + dx * index,
            this.rect.y + dy * index
        ];
    }

    getAttackingIndex (card) {
        for (let i = 0; i < this.countAttacks(); ++i) {
            if (this.pairs[i][0] === card) {
                return i;
            }
        }
    }

    getDefendingOffset (card) {
        const [dx, dy] = this.getOffsetStep(this.countAttacks());
        const index = this.getDefendingIndex(card);
        return [
            this.rect.x + dx * index + this.defenseRect.x,
            this.rect.y + dy * index + this.defenseRect.y
        ];
    }

    getDefendingIndex (card) {
        for (let i = 0; i < this.countAttacks(); ++i) {
            if (this.pairs[i][1] === card) {
                return i;
            }
        }
    }

    getOffsetStep (total) {
        const cardWidth = this.play.getCardWidth();
        return [Club.getOffsetStep(this.rect.w, cardWidth, total, this.constructor.MAX_CARD_SPACE), 0];
    }

    arrange (addition = 0) {
        this.rect = this.play.getElementRect('.table');
        this.defenseRect = this.play.getElementRect('.table-defense');
        const total = this.countAttacks();
        const [dx, dy] = this.getOffsetStep(total + addition);
        for (let i = 0; i < total; ++i) {
            const [attacking, defending] = this.pairs[i];
            const x = this.rect.x + dx * i;
            const y = this.rect.y + dy * i;
            attacking?.setOffset(x, y);
            defending?.setOffset(x + this.defenseRect.x, y + this.defenseRect.y);
        }
        this.reorder();
    }

    reorder () {
        let prev = null;
        for (const [attacking, defending] of this.pairs) {
            if (prev) {
                prev.after(attacking);
            }
            if (defending) {
                attacking.after(defending);
            }
            prev = defending || attacking;
        }
    }
};